# Copyright (c) 2012, Found IT A/S and Piped Project Contributors.
# See LICENSE for details.
import copy
import logging

from cyclone import web
from zope import interface
from twisted.application import strports
from twisted.python import reflect, filepath

from piped import resource, exceptions
from piped_cyclone import handlers


logger = logging.getLogger(__name__)


class InvalidHandlerError(exceptions.PipedError):
    pass


class CycloneProvider(object):
    """ Provides support for running `Cyclone <http://cyclone.io>`_ applications within Piped.

    For more in-depth documentation about Cyclone see: `Cyclone on GitHub <http://github.com/fiorix/cyclone>`_,
    and the `Tornado documentation <http://www.tornadoweb.org/documentation/index.html>`_.

    Configuration example:

    .. code-block:: yaml

        cyclone:
            site_name:
                listen: 8888 # (default)
                application_factory: cyclone.web.Application # (default) - a fully qualified name of a subclass
                application:
                    handlers:
                        # a tuple: pattern, handler, kwargs (optional), name (optional)
                        - ['/pattern', name.of.RequestHandler]
                        # a dict
                        - pattern: /another/(?P<pattern>.*)
                          handler: name.of.RequestHandler
                          name: handler_name # optional
                          kwargs: # optional
                            foo: 123

                    debug: true # see the debugging section below.
                    debug_allow: # optional list of ip addresses that are allowed to see tracebacks and the interactive debugger.
                        - 127.0.0.1
                    debug_timeout: 60 # (default), time in seconds to keep debuggers resident in memory before garbage collecting them.
                    debug_template: debugger.html  # name of the debugger template to use. if not set, uses the built-in template.

    The ``listen`` parameter is a Twisted ``strport``, which can be used to declare which interfaces the server should be listening on, SSL
    parameters and more. For more information, see `strports in the Twisted documentation
    <http://twistedmatrix.com/documents/11.0.0/api/twisted.internet.endpoints.html#serverFromString>`_.

    Each ``handler`` may be one of the following:

        * Fully qualified class name of a RequestHandler instance. The class will be invoked with
          cls.configure(runtime_environment) once if the method is defined. This allows for the
          class to perform any necessary setup / bootstrapping into the runtime_environment.

        * A dict with a ``class`` key set: Same as the above.

        * A dict with the ``provider`` key set. The provider is assumed to provide a subclass of
          RequestHandler as the resource.

          If the provider starts with ``pipeline.``, the provided resource will not be used as
          a RequestHandler, but will be called with ``resource.__call__(baton)``. The baton contains
          the following keys:

            * handler: A RequestHandler that is handling the request.
            * args: A tuple of arguments from the url pattern
            * kwargs: A dict of keyword arguments from the url pattern

          The pipeline should take care of calling ``baton['handler'].finish()`` when the request is
          finished.

    If the application setting ``ui_modules`` is a string, it will be loaded to a python object via
    :meth:`twisted.python.reflect.namedAny`. If it is a dict, all values are assumed to be strings
    that are fully qualified name of :class:`cyclone.web.UIModule` classes.

    Application settings that end with ``_path`` and are :class:`twisted.python.filepath.FilePath` instances
    will be converted to absolute paths. This enables the use of the :ref:`\!path <path-constructor>`-constructor
    in the configuration files.

    Global dependencies can be added to the application settings under the ``piped_dependencies`` key, which
    should be a dict. This dict will be converted to a :class:`~piped.dependencies.DependencyMap` before being
    given to the :class:`cyclone.web.Application` instance as a setting.

    In order to enable debugging tracebacks from the web server, configure the cyclone application as in the example
    configuration above and make sure your request handlers subclasses :class:`piped_cyclone.handlers.DebuggableHandler`.

    .. warning:: Enabling debugging enables allowed clients to execute arbitrary code as the user running :program:`piped`.

    To make stack traces keep their frame information, use :option:`piped -D <piped -D>` when launching :program:`piped`.

    """
    interface.classProvides(resource.IResourceProvider)

    configuration_key = 'cyclone'

    def __init__(self):
        self._configured_handler_factory_ids = set()
        self._applications_by_name = dict()

    def configure(self, runtime_environment):
        self.runtime_environment = runtime_environment
        self.dependency_manager = runtime_environment.dependency_manager
        self.resource_manager = runtime_environment.resource_manager

        cm = runtime_environment.configuration_manager

        for name, config in cm.get(self.configuration_key, dict()).items():
            if not config.get('enabled', True):
                logger.info('Not starting cyclone site [{0}] because it is disabled.'.format(name))
                continue

            # take a deep copy of the configurations, since we're going to modify it
            config = copy.deepcopy(config)

            listen = str(config.pop('listen', '8888'))
            application_factory_name = config.pop('application_factory', 'cyclone.web.Application')
            application_factory = reflect.namedAny(application_factory_name)

            application_config = config.pop('application', dict())

            handlers = application_config.pop('handlers', [])

            transforms = self._resolve_transforms(application_config.pop('transforms', list()))

            if 'ui_modules' in application_config:
                ui_modules = application_config['ui_modules']
                if isinstance(ui_modules, dict):
                    for key, value in ui_modules.items():
                        ui_modules[key] = reflect.namedAny(value)
                elif isinstance(ui_modules, basestring):
                    application_config['ui_modules'] = reflect.namedAny(ui_modules)
                else:
                    raise Exception('Unknown ui_modules type: {0}'.format(type(ui_modules)))

            for key, value in application_config.items():
                if key.endswith('_path') and isinstance(value, filepath.FilePath):
                    application_config[key] = value.path

            for i, handler in enumerate(handlers):
                handlers[i] = self._resolve_urlspec(name, handler)

            if 'piped_dependencies' in application_config:
                for key, value in application_config['piped_dependencies'].items():
                    application_config['piped_dependencies'][key] = dict(provider=value) if isinstance(value, basestring) else value
                application_config['piped_dependencies'] = self.dependency_manager.create_dependency_map(self, **application_config['piped_dependencies'])

            application = application_factory(handlers, transforms=transforms, **application_config)

            service = strports.service(listen, application)
            service.setServiceParent(runtime_environment.application)

            self._applications_by_name[name] = application

            self.resource_manager.register('cyclone.application.{0}'.format(name), provider=self)

    def add_consumer(self, resource_dependency):
        name = resource_dependency.provider.rsplit('.', 1)[-1]
        application = self._applications_by_name[name]
        resource_dependency.on_resource_ready(application)

    def _resolve_urlspec(self, site_name, url_spec):
        if isinstance(url_spec, (list, tuple)):
            url_spec = dict(zip(['pattern', 'handler', 'kwargs', 'name'], url_spec))

        if not 'pattern' in url_spec or not 'handler' in url_spec:
            msg = 'Missing attributes in handler configuration in site [{0}]: [{1}]'.format(site_name, url_spec)
            hint = 'Both "pattern" and "handler" must be defined.'
            raise InvalidHandlerError(msg, hint)

        # string -> class specification
        if isinstance(url_spec['handler'], basestring):
            url_spec['handler'] = {'class':url_spec['handler']}

        if 'provider' in url_spec['handler']:
            handler = url_spec.pop('handler')
            provider = handler['provider']
            dependency = self.dependency_manager.add_dependency(self, handler)
            url_spec.setdefault('kwargs', dict()).update(dependency=dependency)

            if provider.startswith('pipeline.'):
                url_spec['handler_class'] = handlers.PipelineRequestHandler
                return web.url(**url_spec)

            url_spec['handler_class'] = handlers.PipedRequestHandlerProxy(dependency)
            return web.url(**url_spec)

        if 'class' in url_spec['handler']:
            handler = url_spec.pop('handler')

            cls = reflect.namedAny(handler['class'])
            if id(cls) not in self._configured_handler_factory_ids:
                self._configured_handler_factory_ids.add(id(cls))

                try:
                    cls.configure(self.runtime_environment)
                except AttributeError as ae:
                    if "has no attribute 'configure'" not in ae.args[0]:
                        raise

            url_spec['handler_class'] = cls
            return web.url(**url_spec)

        msg = 'Invalid handler configuration in site [{0}]: [{1}]'.format(site_name, url_spec)
        raise InvalidHandlerError(msg)

    def _resolve_transforms(self, transforms):
        if not transforms:
            return None

        for i, transform in transforms:
            transforms[i] = reflect.namedAny(transform)

        return transforms



# Copyright (c) 2012, Found IT A/S and Piped Project Contributors.
# See LICENSE for details.
import logging
import os
import json

from cyclone import web
from zope import interface
from twisted.internet import defer, reactor
from twisted.web import util
from twisted.python import failure

from piped import debugger


logger = logging.getLogger(__name__)


class DebuggableHandler(web.RequestHandler):
    """ A request handler that lets allowed users view full tracebacks and execute
    code interactively to inspect the traceback frames.

    Subclasses must make sure that :meth:`.prepare()` is called. If you want to require login
    in addition to the built-in ip-based checks, subclass this class and override
    :meth:`~DebuggableHandler.post_debug` and :meth:`~DebuggableHandler.get_debug`::

        class AuthingDebuggableHandler(web.RequestHandler):
            @web.authenticated
            def post_debug(self, *a, **kw):
                super(AuthingDebuggableHandler, self).post_debug(*a, **kw)

            @web.authenticated
            def get_debug(self, *a, **kw):
                super(AuthingDebuggableHandler, self).get_debug(*a, **kw)
    """
    _debugging_failures = dict()
    _debugging_timeouts = dict()

    def prepare(self):
        """ Override this request handlers methods if the user tries to access a debugger. """
        if self.settings.get('debug', False) and self.get_argument('__debug__', None):
            self.post = self.post_debug
            self.get = self.get_debug

    def post_debug(self, *a, **kw):
        """ Executes an expression within a debugger frame. """
        debugger = self.get_debugger(self.get_argument('__debug__', None), timeout=self.settings.get('debug_timeout', 60))
        if not debugger:
            raise web.HTTPError(404, 'No such debugger.')

        allowed = self.settings.get('debug_allow', list())

        if not self.request.remote_ip in allowed:
            raise web.HTTPError(403)

        expr = self.get_argument('expr', None)
        frame_no = self.get_argument('frame_no', None)

        result = debugger.exec_expr(expr, int(frame_no))
        self.finish(json.dumps(result))

    def get_debug(self, *a, **kw):
        """ Returns a HTML-page that serves as a front-end to the debugger. """
        debugger = self.get_debugger(self.get_argument('__debug__', None), timeout=self.settings.get('debug_timeout', 60))

        if not debugger or not self.request.remote_ip in self.settings.get('debug_allow', list()):
            raise web.HTTPError(404, 'No such debugger.')

        self.finish(self._get_debug_html(debugger.failure))

    def write_error(self, status_code, **kwargs):
        if not self.settings.get('debug', False):
            super(DebuggableHandler, self).write_error(status_code, **kwargs)
            return

        # store the current failure reason
        reason = None

        if 'exception' in kwargs and isinstance(kwargs['exception'], failure.Failure):
            reason = kwargs['exception']

        if 'exc_info' in kwargs:
            try:
                raise exc_info[0], exc_info[1], exc_info[2]
            except Exception as e:
                reason = failure.Failure()

        if reason:
            self.register_failure(reason, timeout=self.settings.get('debug_timeout', 60))

            ajax_endpoint = self.request.path + '?__debug__='+str(id(reason))
            full_url_to_debugger = self.request.protocol + '://' + self.request.host + ajax_endpoint

            logger.error('Debugger for [{0}] started at [{1}]'.format(reason.getErrorMessage(), full_url_to_debugger))

            self.request.arguments['__debug__'] = [str(id(reason))]
            if self.request.remote_ip in self.settings.get('debug_allow', list()):
                return self.get_debug()

        super(DebuggableHandler, self).write_error(status_code, **kwargs)

    def _get_debug_html(self, reason):
        traceback_as_html = util.formatFailure(reason)
        ajax_endpoint = self.request.path + '?__debug__='+str(id(reason))

        here = os.path.abspath(os.path.dirname(__file__))

        template_name = self.settings.get('debug_template', os.path.join(here, 'templates', 'debugger.html'))

        try:
            return self.render_string(template_name, traceback_as_html=traceback_as_html, ajax_endpoint=ajax_endpoint)
        except Exception as e:
            logger.error('Error while rendering debug html template.', exc_info=True)
            raise e

    @classmethod
    def register_failure(cls, reason, timeout):
        key = str(id(reason))
        cls._debugging_failures[key] = debugger.Debugger(reason)

        cls._debugging_timeouts[key] = reactor.callLater(timeout, cls._timeout_debugger, key)

    @classmethod
    def get_debugger(cls, id, timeout):
        debugger = cls._debugging_failures.get(id, None)
        delayed_call = cls._debugging_timeouts.pop(id, None)

        if delayed_call:
            delayed_call.cancel()
            cls._debugging_timeouts.pop(id, None)
            cls._debugging_timeouts[id] = reactor.callLater(timeout, cls._timeout_debugger, id)

        return debugger

    @classmethod
    def _timeout_debugger(cls, id):
        cls._debugging_failures.pop(id, None)
        cls._debugging_timeouts.pop(id, None)


class PipedRequestHandlerProxy(object):
    """ A simple proxy that uses a dependency to process web requests.

    Assumes the provided resource acts like a :class:`cyclone.web.RequestHandler`.
    """

    def __init__(self, dependency):
        self.dependency = dependency

    def __call__(self, *args, **kwargs):
        self.handler_args = args
        self.handler_kwargs = kwargs
        return self

    @defer.inlineCallbacks
    def _execute(self, *a, **kw):
        handler_class = yield self.dependency.wait_for_resource()
        handler = handler_class(*self.handler_args, **self.handler_kwargs)
        yield handler._execute(*a, **kw)


class PipelineRequestHandler(DebuggableHandler):
    """ Request handler for requests that will be served through pipelines. """

    def __init__(self, *args, **kwargs):
        super(PipelineRequestHandler, self).__init__(*args, **kwargs)

        for method in self.SUPPORTED_METHODS:
            setattr(self, method.lower(), self.handle)

    def initialize(self, dependency=None):
        self.dependency = dependency

    @web.asynchronous
    @defer.inlineCallbacks
    def handle(self, *args, **kwargs):
        pipeline = yield self.dependency.wait_for_resource()
        pipeline(dict(handler=self, args=args, kwargs=kwargs))


# Copyright (c) 2012, Found IT A/S and Piped Project Contributors.
# See LICENSE for details.
import re
import socket

import mock
from cyclone import web, httpclient
from twisted.application import service
from twisted.internet import defer
from twisted.python import filepath, reflect
from twisted.trial import unittest
from piped import processing
from piped.dependencies import ResourceDependency

from piped_cyclone import providers, handlers


class TestHelloHandler(web.RequestHandler):
    def get(self):
        self.finish('get from test hello handler')


class TestRaisingHandler(handlers.DebuggableHandler):
    def get(self):
        foo = 123
        1/0

class FooModule(web.UIModule):
    pass


class TestApplication(web.Application):
    pass


class CycloneProviderTest(unittest.TestCase):

    def setUp(self):
        self.runtime_environment = processing.RuntimeEnvironment()
        self.service = service.IService(self.runtime_environment.application)
        self.dependency_manager = self.runtime_environment.dependency_manager
        self.configuration_manager = self.runtime_environment.configuration_manager
        self.resource_manager = self.runtime_environment.resource_manager

        self.dependency_manager.configure(self.runtime_environment)

    def tearDown(self):
        if self.service.running:
            self.service.stopService()

    def get_free_port(self):
        sock = socket.socket()
        sock.bind(('', 0))
        free_port = sock.getsockname()[1]
        sock.close()
        return free_port

    def get_configured_cyclone(self, name='test_cyclone', **config):
        if 'listen' not in config:
            config['listen'] = self.get_free_port()

        resource_name = 'cyclone.application.{0}'.format(name)
        self.configuration_manager.set('cyclone.{0}'.format(name), config)

        provider = providers.CycloneProvider()
        provider.configure(self.runtime_environment)

        provider = self.resource_manager.get_provider_or_fail(resource_name)
        dependency = ResourceDependency(provider=resource_name)
        provider.add_consumer(dependency)

        return dependency.get_resource()

    def test_invalid_handlers(self):
        invalid_handler_configs = [
            dict(foo='bar'), # missing "handler" and "pattern"
            dict(handler='bar'), # missing "pattern"
            dict(pattern='bar'), # missing "handler"
            dict(handler=set()) # unknown handler type
        ]

        for invalid_handler_config in invalid_handler_configs:
            self.assertRaises(providers.InvalidHandlerError, self.get_configured_cyclone,
                listen=8888,
                application = dict(
                    handlers = [
                        invalid_handler_config
                    ]
                )
            )

    @defer.inlineCallbacks
    def test_simple(self):
        port = self.get_free_port()
        app = self.get_configured_cyclone(
            listen = port,
            application=dict(
                handlers=[
                    ('/', reflect.fullyQualifiedName(TestHelloHandler))
                ],

                templates_path = filepath.FilePath('.'),
                static_path = filepath.FilePath('.'),

                ui_modules = dict(
                    foo = reflect.fullyQualifiedName(FooModule)
                )
            )
        )
        self.assertIsInstance(app, web.Application)


        # the _paths in the setting should have been rewritten from FilePaths to strings
        self.assertIsInstance(app.settings.templates_path, basestring)
        self.assertIsInstance(app.settings.static_path, basestring)

        # the foo module should be resolved to an instance
        self.assertIdentical(app.settings.ui_modules['foo'], FooModule)

        # start the web application
        yield self.service.startService()

        response = yield httpclient.fetch('http://localhost:{0}'.format(port))

        self.assertEqual(response.body, 'get from test hello handler')
        self.assertEqual(response.code, 200)

    def test_custom_app(self):
        app = self.get_configured_cyclone(
            application_factory = reflect.fullyQualifiedName(TestApplication)
        )
        self.assertIsInstance(app, TestApplication)

    @defer.inlineCallbacks
    def test_with_pipeline(self):
        class Provider:
            batons = list()

            def add_consumer(self, resource_dependency):
                resource_dependency.on_resource_ready(self)

            def __call__(self, baton):
                self.batons.append(baton)
                baton['handler'].set_status(201)
                baton['handler'].finish('hello from the pipeline')

        provider = Provider()
        self.resource_manager.register('pipeline.foo', provider)

        port = self.get_free_port()
        app = self.get_configured_cyclone(
            listen = port,
            application=dict(handlers=[
                ('/pipeline/?(?P<rest>.*)', dict(provider='pipeline.foo'))
            ])
        )
        self.dependency_manager.resolve_initial_states()

        # start the web application
        yield self.service.startService()

        response = yield httpclient.fetch('http://localhost:{0}/pipeline/123'.format(port))

        self.assertEqual(response.body, 'hello from the pipeline')
        self.assertEqual(response.code, 201)

        self.assertEqual(len(provider.batons), 1)
        self.assertIsInstance(provider.batons[0]['handler'], handlers.PipelineRequestHandler)

        self.assertEqual(provider.batons[0]['kwargs'], dict(rest=u'123'))

    @defer.inlineCallbacks
    def test_with_dependency(self):
        class Provider(web.RequestHandler):
            @classmethod
            def add_consumer(cls, resource_dependency):
                resource_dependency.on_resource_ready(cls)

            def get(self):
                self.set_status(202)
                self.finish('hello from the foo provider')

        self.resource_manager.register('a.dependency.foo', Provider)

        port = self.get_free_port()
        app = self.get_configured_cyclone(
            listen = port,
            application=dict(
                handlers=[
                    ('/dependency', dict(provider='a.dependency.foo'))
                ],
                piped_dependencies = dict(
                    foo = 'a.dependency.foo'
                )
            )
        )
        self.dependency_manager.resolve_initial_states()

        # assert that the piped_dependencies is available in the application settings
        self.assertEqual(app.settings.piped_dependencies.foo, Provider)

        # start the web application
        yield self.service.startService()

        response = yield httpclient.fetch('http://localhost:{0}/dependency'.format(port))

        self.assertEqual(response.body, 'hello from the foo provider')
        self.assertEqual(response.code, 202)

    @defer.inlineCallbacks
    def test_debugging(self):
        port = self.get_free_port()
        app = self.get_configured_cyclone(
            listen = port,
            application=dict(
                handlers=[
                    ('/', reflect.fullyQualifiedName(TestRaisingHandler))
                ],
                debug = True,
                debug_allow = ['127.0.0.1'],
                debug_timeout = 10
            )
        )

        # start the web application
        yield self.service.startService()

        with mock.patch.object(handlers, 'logger') as mocked_logger:
            try:
                # perform the web request with debugging enabled so we can inspect the stack :)
                defer.setDebugging(True)

                response = yield httpclient.fetch('http://localhost:{0}/'.format(port))
            finally:
                defer.setDebugging(False)

        self.assertEqual(response.code, 500)
        self.assertIn('web.Application Traceback (most recent call last)', response.body)
        self.assertIn('exceptions.ZeroDivisionError', response.body)

        debugger_id = re.compile(r'.*__debug__=(?P<id>.*)".*').search(response.body).groupdict()['id']

        # in the context, "foo" has been repr'ed to a string by failure.Failure.cleanFailure(). this does not
        # happen if piped -D is used, which replaces the failure class with a subclass that skips the cleaning.
        #
        # exec "foo + ''.join(reversed(foo))", which should return 123321 as a json-encoded string
        exec_response = yield httpclient.fetch('http://localhost:{0}/?__debug__={1}&frame_no=-1&expr=foo%2B"".join(reversed(foo))'.format(port, debugger_id), method='POST')
        self.assertEqual(exec_response.body, '"\'123321\'\\n"')

        # clean out the debugger timeout, so the test does not result in an unclean reactor. this is
        # a lot faster than waiting for the debug_timeout to occur
        delayed_call = TestRaisingHandler._debugging_timeouts.pop(debugger_id, None)
        if delayed_call:
            delayed_call.cancel()

    @defer.inlineCallbacks
    def test_debugging_allow_list(self):
        port = self.get_free_port()
        app = self.get_configured_cyclone(
            listen = port,
            application=dict(
                handlers=[
                    ('/', reflect.fullyQualifiedName(TestRaisingHandler))
                ],
                debug = True,
                debug_allow = [],
                debug_timeout = 0 # setting the timeout to 0 means the debugger will disappear right after the request is done.
            )
        )

        # start the web application
        yield self.service.startService()

        with mock.patch.object(handlers, 'logger') as mocked_logger:
            response = yield httpclient.fetch('http://localhost:{0}/'.format(port))

            # we should get an error..
            self.assertEqual(response.code, 500)
            # .. but should be unable to debug because we're not in the debug_allow list.
            self.assertNotIn('web.Application Traceback (most recent call last)', response.body)
            self.assertIn('500: Internal Server Error', response.body)